;	include	<p16F877A.inc>
	include <p18F442.inc>

TestVal	equ	1234567890

PIC18	equ	1
ROUND	equ	1

;variables
	cblock	0x20
		root 	: 4
		trial 	: 4
		arg 	: 4
	endc
	
;macros
clrf16	macro	var		; clear a 16-bit variable
	clrf	var	
	clrf	var+1	
	endm

clrf32	macro	var		; clear a 32-bit variable
	clrf16	var	
	clrf16	var+2	
	endm
	
ior32	macro	var1,var2	; inclusive-or 2 32-bit values
	movf	var2,w
	iorwf	var1,f
	movf	var2+1,w
	iorwf	var1+1,f
	movf	var2+2,w
	iorwf	var1+2,f
	movf	var2+3,w
	iorwf	var1+3,f
	endm

xor32	macro	var1,var2	; exclusive-or 2 32-bit values
	movf	var2,w
	xorwf	var1,f
	movf	var2+1,w
	xorwf	var1+1,f
	movf	var2+2,w
	xorwf	var1+2,f
	movf	var2+3,w
	xorwf	var1+3,f
	endm

#if	PIC18

cmp32	macro	var1,var2	; compare 2 32-bit values
	movf	var2,w	
	subwf	var1,w		; subtract, but discard result
	movf	var2+1,w
	subwfb	var1+1,w	; subtract with borrow for the rest
	movf	var2+2,w
	subwfb	var1+2,w
	movf	var2+3,w
	subwfb	var1+3,w	; result ignored, but C flag set if var1<var2
	endm

cmp16	macro	var1,var2	; compare 2 16-bit values
	movf	var2,w	
	subwf	var1,w		; subtract, but discard result
	movf	var2+1,w
	subwfb	var1+1,w	; subtract with borrow, C flag has result
	endm

inc24	macro	var		; increment a 24-bit variable
	local	IncDone
	incfsz	var,f		; increment, skip branch if 0
	bra	IncDone		; if non-zero, done with increment
	infsnz	var+1,f		; increment 2nd byte, skip if non-zero
	incf	var+2,f		; only if both were 0 above
IncDone:
	endm

;rrf32 is not different from the PIC16 version, just the mnemonic was changed
rrf32	macro	var		; perform right shift on 32-bit variable
	rrcf	var+3,f
	rrcf	var+2,f
	rrcf	var+1,f
	rrcf	var,f
	endm

add32	macro	var1,var2	; add two 32-bit variables
	movf	var2,w	
	addwf	var1,f	
	movf	var2+1,w
	addwfc	var1+1,f	; add with carry for the rest
	movf	var2+2,w
	addwfc	var1+2,f
	movf	var2+3,w
	addwfc	var1+3,f
	endm

sub32	macro	var1,var2	; two 32-bit variables
	movf	var2,w	
	subwf	var1,f	
	movf	var2+1,w
	subwfb	var1+1,f	; subtract with borrow for the rest
	movf	var2+2,w
	subwfb	var1+2,f
	movf	var2+3,w
	subwfb	var1+3,f
	endm
	
#else	; PIC16 versions below

bra	macro	targ		; branch, for compatibility with PIC18
	goto	targ
	endm
	
cmp32	macro	var1,var2	; compare 2 32-bit values
	movf	var2,w	
	subwf	var1,w		; subtract but discard result
	movf	var2+1,w
	btfss	STATUS,C	; C flag clear means borrow
	incfsz	var2+1,w	; borrow by increasing subtrahend
	subwf	var1+1,w	; skip subtract if zero to preserve C flag
	movf	var2+2,w
	btfss	STATUS,C
	incfsz	var2+2,w
	subwf	var1+2,w
	movf	var2+3,w
	btfss	STATUS,C
	incfsz	var2+3,w
	subwf	var1+3,w
	endm

cmp16	macro	var1,var2	; compare 2 32-bit values
	movf	var2,w	
	subwf	var1,w		; subtract but discard result
	movf	var2+1,w
	btfss	STATUS,C	; C flag clear means borrow
	incfsz	var2+1,w	; borrow by increasing subtrahend
	subwf	var1+1,w	; skip subtract if zero to preserve C flag
	endm

inc24	macro	var		; increment a 24-bit variable
	local	IncDone
	incfsz	var,f		; increment, skip branch if 0
	bra	IncDone		; if non-zero, done with increment
	incfsz	var+1,f		; increment 2nd byte, skip branch if 0
	bra	IncDone		; if non-zero, done with increment
	incf	var+2,f		; only if both were 0 above
IncDone:
	endm

;rrf32 is not different from the PIC18 version, just the mnemonic was changed
rrf32	macro	var		; perform right shift on 32-bit variable
	rrf	var+3,f
	rrf	var+2,f
	rrf	var+1,f
	rrf	var,f
	endm

add32	macro	var1,var2	; add two 32-bit variables
	movf	var2,w	
	addwf	var1,f	
	movf	var2+1,w
	btfsc	STATUS,C	; C flag set means carry
	incfsz	var2+1,w	; carry by increasing addend
	addwf	var1+1,f	; skip add if zero to preserve C flag
	movf	var2+2,w
	btfsc	STATUS,C
	incfsz	var2+2,w
	addwf	var1+2,f
	movf	var2+3,w
	btfsc	STATUS,C
	incfsz	var2+3,w
	addwf	var1+3,f
	endm

sub32	macro	var1,var2	; subtract two 32-bit variables
	movf	var2,w	
	subwf	var1,f	
	movf	var2+1,w
	btfss	STATUS,C	; C flag clear means borrow
	incfsz	var2+1,w	; borrow by increasing subtrahend
	subwf	var1+1,f	; skip subtract if zero to preserve C flag
	movf	var2+2,w
	btfss	STATUS,C
	incfsz	var2+2,w
	subwf	var1+2,f
	movf	var2+3,w
	btfss	STATUS,C
	incfsz	var2+3,w
	subwf	var1+3,f
	endm

#endif	; else PIC18

; End of macros

;**********************************************************************
;
; Test driver
;

	org	0
TestSqrt:
; Copy test value into arg
	movlw	TestVal & 0xFF
	movwf	arg
	movlw	(TestVal >> 8) & 0xFF
	movwf	arg+1
	movlw	(TestVal >> 16) & 0xFF
	movwf	arg+2
	movlw	(TestVal >> 24) & 0xFF
	movwf	arg+3
; arg is loaded with test value
	call	SquareRoot	; call our square root function
Wait:	bra	Wait

;**********************************************************************
;
; Fast square root function
;

SquareRoot:
	clrf32	root		; root = 0
	clrf16	trial		; trial = 0x8000
	clrf	trial+2		;  "
	movlw	0x80		;  "
	movwf	trial+3		;  "
	bcf	STATUS,C	; Clear C flag for following rotate
RootBit:
	rrf32	trial		; trial >>= 1
	ior32	root,trial	; root != trial
	cmp32	root,arg	; if (arg < root) - C flag set if arg < root
	bnc	RootFits	;  "
; Didn't fit, remove trial bit
	xor32	root,trial	; root ^= trial
	bra	ShiftRight

RootFits:
	sub32	arg,root	; arg -= root
	add32	root,trial	; root += trial
ShiftRight:
	bcf	STATUS,C	; Clear C flag for following rotate
	rrf32	root		; root >>= 1 - Always leaves C flag clear
	rrf32	trial		; trial >>= 1
	bnc	RootBit		; if trial bit shifted out, we're done

#if	ROUND
; Compute rounding. Note that root always fits in 16 bits here,
; but it could round up to 17 bits.
	movf	arg+2,f		; see if bits 17-23 are non-zero
	bnz	IncResult	; if non-zero, arg is big, always > root
	cmp16	arg,root	; if (arg > root) - C flag set if root < arg
	bnc	Done		;  "
IncResult:
	inc24	root		; root++
Done:
#endif	;ROUND

	retlw	0

	end
